/*
 * kgraft_patch_bsc971793
 *
 * Fix for CVE-2016-3134, bsc#971793, arp_tables module
 * Fix for CVE-2016-4997, bsc#986377
 *
 *  Upstream commits:
 *  bdf533de6968 ("netfilter: x_tables: validate e->target_offset early")
 *  6e94e0cfb088 ("netfilter: x_tables: make sure e->next_offset covers remaining blob size")
 *  54d83fc74aa9 ("netfilter: x_tables: fix unconditional helper")
 *
 *  ce683e5f9d04 ("netfilter: x_tables: check for bogus target offset")
 *
 *  SLE12 commits:
 *  bf636090e574fbf6f6262f9aba2a37b01ab90f6e
 *  04ead165cb899415697570588f14e74bec5f7957
 *  7f167def8c00bc3b727fa3eebf886f901b1c748c
 *
 *  c432d109f6a4f68a92a623e3968e8d51f5a47133
 *  ... (>10 patches in the series)
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/netfilter_arp/arp_tables.h>
#include <net/sock.h>

#include "local_arp.h"
#include "../bsc986377/kgr_patch_bsc986377_xtables.h"

#define KGR_PATCHED_MODULE      "arp_tables"

static int (*kgr_do_add_counters)(struct net *net, const void __user *user, unsigned int len, int compat);
static int (*kgr__do_replace)(struct net *net, const char *name, unsigned int valid_hooks, struct xt_table_info *newinfo, unsigned int num_counters, void __user *counters_ptr);
static void (*kgr_xt_compat_lock)(u_int8_t af);
static int (*kgr_xt_check_target)(struct xt_tgchk_param *par, unsigned int size, u_int8_t proto, bool inv_proto);
static int (*kgr_xt_compat_target_offset)(const struct xt_target *target);
static void (*kgr_xt_compat_flush_offsets)(u_int8_t af);
static void (*kgr_xt_compat_init_offsets)(u_int8_t af, unsigned int number);
static int (*kgr_xt_compat_add_offset)(u_int8_t af, unsigned int offset, int delta);
static struct xt_target *(*kgr_xt_request_find_target)(u8 af, const char *name, u8 revision);
static struct xt_table_info *(*kgr_xt_alloc_table_info)(unsigned int size);
static void (*kgr_xt_free_table_info)(struct xt_table_info *info);
static void (*kgr_xt_compat_unlock)(u_int8_t af);
struct xt_table *(*kgr_xt_register_table)(struct net *net, const struct xt_table *input_table, struct xt_table_info *bootstrap, struct xt_table_info *newinfo);
static seqcount_t *kgr_xt_recseq;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "arp_tables:do_add_counters", (void *)&kgr_do_add_counters },
	{ "arp_tables:__do_replace", (void *)&kgr__do_replace },
	{ "xt_compat_lock", (void *)&kgr_xt_compat_lock },
	{ "xt_check_target", (void *)&kgr_xt_check_target },
	{ "xt_compat_target_offset", (void *)&kgr_xt_compat_target_offset },
	{ "xt_compat_flush_offsets", (void *)&kgr_xt_compat_flush_offsets },
	{ "xt_compat_init_offsets", (void *)&kgr_xt_compat_init_offsets },
	{ "xt_compat_add_offset", (void *)&kgr_xt_compat_add_offset },
	{ "xt_request_find_target", (void *)&kgr_xt_request_find_target },
	{ "xt_alloc_table_info", (void *)&kgr_xt_alloc_table_info },
	{ "xt_free_table_info", (void *)&kgr_xt_free_table_info },
	{ "xt_compat_unlock", (void *)&kgr_xt_compat_unlock },
	{ "xt_register_table", (void *)&kgr_xt_register_table },
	{ "xt_recseq", (void *)&kgr_xt_recseq },
};

/* inlined */
static inline const struct xt_entry_target *
kgr_arpt_get_target_c(const struct arpt_entry *e)
{
	return arpt_get_target((struct arpt_entry *)e);
}

/* inlined */
/*
 * Fix CVE-2016-3134
 *  -static inline bool kgr_unconditional(const struct arpt_arp *arp)
 *  +next row
 */
static inline bool kgr_unconditional(const struct arpt_entry *e)
{
	static const struct arpt_arp uncond;

	/*
	 * Fix CVE-2016-3134
	 *  -return memcmp(arp, &uncond, sizeof(uncond)) == 0;
	 *  +next 2 rows
	 */
	return e->target_offset == sizeof(struct arpt_entry) &&
		memcmp(&e->arp, &uncond, sizeof(uncond)) == 0;
}

/* inlined */
static bool kgr_check_underflow(const struct arpt_entry *e)
{
	const struct xt_entry_target *t;
	unsigned int verdict;

	/*
	 * Fix CVE-2016-3134	
	 *  -if (!kgr_unconditional(&e->arp))
	 *  +next row
	 */
	if (!kgr_unconditional(e))
		return false;
	t = kgr_arpt_get_target_c(e);
	if (strcmp(t->u.user.name, XT_STANDARD_TARGET) != 0)
		return false;
	verdict = ((struct xt_standard_target *)t)->verdict;
	verdict = -verdict - 1;
	return verdict == NF_DROP || verdict == NF_ACCEPT;
}

/* inlined */
static inline int kgr_arp_checkentry(const struct arpt_arp *arp)
{
	if (arp->flags & ~ARPT_F_MASK) {
		duprintf("Unknown flag bits set: %08X\n",
			 arp->flags & ~ARPT_F_MASK);
		return 0;
	}
	if (arp->invflags & ~ARPT_INV_MASK) {
		duprintf("Unknown invflag bits set: %08X\n",
			 arp->invflags & ~ARPT_INV_MASK);
		return 0;
	}

	return 1;
}

/* inlined */
/*
 * Fix CVE-2016-3134
 *  -static inline int kgr_check_entry(const struct arpt_entry *e, const char *name)
 *  +next row
 */
/*
 * Fix CVE-2016-4997
 *  function removed
 */
//static inline int kgr_check_entry(const struct arpt_entry *e)
//{
//	const struct xt_entry_target *t;
//
//	/*
//	 * Fix CVE-2016-3134
//	 *  -if (!kgr_arp_checkentry(&e->arp)) {
//	 *  -	duprintf("arp_tables: arp check failed %p %s.\n", e, name);
//	 *  +next row
//	 */
//	if (!kgr_arp_checkentry(&e->arp))
//		return -EINVAL;
//	/*
//	 * -}
//	 */
//
//	if (e->target_offset + sizeof(struct xt_entry_target) > e->next_offset)
//		return -EINVAL;
//
//	t = kgr_arpt_get_target_c(e);
//	if (e->target_offset + t->u.target_size > e->next_offset)
//		return -EINVAL;
//
//	return 0;
//}

/* inlined */
static inline int kgr_check_entry_size_and_hooks(struct arpt_entry *e,
	struct xt_table_info *newinfo, const unsigned char *base,
	const unsigned char *limit, const unsigned int *hook_entries,
	const unsigned int *underflows, unsigned int valid_hooks)
{
	unsigned int h;
	/*
	 * Fix CVE-2016-3134
	 *  +next row
	 */
	int err;

	if ((unsigned long)e % __alignof__(struct arpt_entry) != 0 ||
	/*
	 * Fix CVE-2016-3134
	 *  -    (unsigned char *)e + sizeof(struct arpt_entry) >= limit) {
	 *  +next 2 rows
	 */
	    (unsigned char *)e + sizeof(struct arpt_entry) >= limit ||
	    (unsigned char *)e + e->next_offset > limit) {
		duprintf("Bad offset %p\n", e);
		return -EINVAL;
	}

	if (e->next_offset
	    < sizeof(struct arpt_entry) + sizeof(struct xt_entry_target)) {
		duprintf("checking: element %p size %u\n",
			 e, e->next_offset);
		return -EINVAL;
	}
	/*
	 * Fix CVE-2016-3134
	 *  +next 4 rows
	 */
	/*
	 * Fix CVE-2016-4997
	 *  -err = kgr_check_entry(e);
	 *  +next 5 lines
	 */
	if (!kgr_arp_checkentry(&e->arp))
		return -EINVAL;

	err = kgr_xt_check_entry_offsets(e, e->elems, e->target_offset,
				     e->next_offset);
	if (err)
		return err;

	/* Check hooks & underflows */
	for (h = 0; h < NF_ARP_NUMHOOKS; h++) {
		if (!(valid_hooks & (1 << h)))
			continue;
		if ((unsigned char *)e - base == hook_entries[h])
			newinfo->hook_entry[h] = hook_entries[h];
		if ((unsigned char *)e - base == underflows[h]) {
			if (!kgr_check_underflow(e)) {
				/*
				 * Fix CVE-2016-3134				
				 *  -pr_err("Underflows must be unconditional and "
				 *  -       "use the STANDARD target with "
				 *  -       "ACCEPT/DROP\n");
				 *  +next 3 rows
				 */
				pr_debug("Underflows must be unconditional and "
					 "use the STANDARD target with "
					 "ACCEPT/DROP\n");
				return -EINVAL;
			}
			newinfo->underflow[h] = underflows[h];
		}
	}

	/* Clear counters and comefrom */
	e->counters = ((struct xt_counters) { 0, 0 });
	e->comefrom = 0;
	return 0;
}

/* to-be-patched function, but ambiguous */
/*
 * Fix CVE-2016-4997
 *  -int kgr_mark_source_chains(const struct xt_table_info *newinfo,
 *  -	unsigned int valid_hooks, void *entry0)
 *  +next 2 lines
 */
int kgr_mark_source_chains(const struct xt_table_info *newinfo,
	unsigned int valid_hooks, void *entry0, unsigned int *offsets)
{
	unsigned int hook;

	/* No recursion; use packet counter to save back ptrs (reset
	 * to 0 as we leave), and comefrom to save source hook bitmask.
	 */
	for (hook = 0; hook < NF_ARP_NUMHOOKS; hook++) {
		unsigned int pos = newinfo->hook_entry[hook];
		struct arpt_entry *e
			= (struct arpt_entry *)(entry0 + pos);

		if (!(valid_hooks & (1 << hook)))
			continue;

		/* Set initial back pointer. */
		e->counters.pcnt = pos;

		for (;;) {
			const struct xt_standard_target *t
				= (void *)kgr_arpt_get_target_c(e);
			int visited = e->comefrom & (1 << hook);

			if (e->comefrom & (1 << NF_ARP_NUMHOOKS)) {
				pr_notice("arptables: loop hook %u pos %u %08X.\n",
				       hook, pos, e->comefrom);
				return 0;
			}
			e->comefrom
				|= ((1 << hook) | (1 << NF_ARP_NUMHOOKS));

			/* Unconditional return/END. */
			/*
			 * Fix CVE-2016-3134
			 *  -if ((e->target_offset == sizeof(struct arpt_entry) &&
			 *  +next row
			 */
			if ((kgr_unconditional(e) &&
			     (strcmp(t->target.u.user.name,
				     XT_STANDARD_TARGET) == 0) &&
			/*
			 * Fix CVE-2016-3134			 
			 *  -  t->verdict < 0 && kgr_unconditional(&e->arp)) ||
			 *  - visited) {
			 *  +next row
			 */
			     t->verdict < 0) || visited) {
				unsigned int oldpos, size;

				if ((strcmp(t->target.u.user.name,
					    XT_STANDARD_TARGET) == 0) &&
				    t->verdict < -NF_MAX_VERDICT - 1) {
					duprintf("mark_source_chains: bad "
						"negative verdict (%i)\n",
								t->verdict);
					return 0;
				}

				/* Return: backtrack through the last
				 * big jump.
				 */
				do {
					e->comefrom ^= (1<<NF_ARP_NUMHOOKS);
					oldpos = pos;
					pos = e->counters.pcnt;
					e->counters.pcnt = 0;

					/* We're at the start. */
					if (pos == oldpos)
						goto next;

					e = (struct arpt_entry *)
						(entry0 + pos);
				} while (oldpos == pos + e->next_offset);

				/* Move along one */
				size = e->next_offset;
				e = (struct arpt_entry *)
					(entry0 + pos + size);
				/*
				 * Fix CVE-2016-4997
				 *  +next 2 lines
				 */
				if (pos + size >= newinfo->size)
					return 0;
				e->counters.pcnt = pos;
				pos += size;
			} else {
				int newpos = t->verdict;

				if (strcmp(t->target.u.user.name,
					   XT_STANDARD_TARGET) == 0 &&
				    newpos >= 0) {
					if (newpos > newinfo->size -
						sizeof(struct arpt_entry)) {
						duprintf("mark_source_chains: "
							"bad verdict (%i)\n",
								newpos);
						return 0;
					}

					/* This a jump; chase it. */
					duprintf("Jump rule %u -> %u\n",
						 pos, newpos);
					/*
					 * Fix CVE-2016-4997
					 *  +next 5 lines
					 */
					if (!kgr_xt_find_jump_offset(offsets, newpos,
								 newinfo->number))
						return 0;
					e = (struct arpt_entry *)
						(entry0 + newpos);
				} else {
					/* ... this is a fallthru */
					newpos = pos + e->next_offset;
					/*
					 * Fix CVE-2016-4997
					 *  +next 2 lines
					 */
					if (newpos >= newinfo->size)
						return 0;
				}
				e = (struct arpt_entry *)
					(entry0 + newpos);
				e->counters.pcnt = pos;
				pos = newpos;
			}
		}
		next:
		duprintf("Finished chain %u\n", hook);
	}
	return 1;
}

/* inlined */
static inline int kgr_check_target(struct arpt_entry *e, const char *name)
{
	struct xt_entry_target *t = arpt_get_target(e);
	int ret;
	struct xt_tgchk_param par = {
		.table     = name,
		.entryinfo = e,
		.target    = t->u.kernel.target,
		.targinfo  = t->data,
		.hook_mask = e->comefrom,
		.family    = NFPROTO_ARP,
	};

	ret = kgr_xt_check_target(&par, t->u.target_size - sizeof(*t), 0, false);
	if (ret < 0) {
		duprintf("arp_tables: check failed for `%s'.\n",
			 t->u.kernel.target->name);
		return ret;
	}
	return 0;
}

/* inlined */
static inline int
kgr_find_check_entry(struct arpt_entry *e, const char *name, unsigned int size)
{
	struct xt_entry_target *t;
	struct xt_target *target;
	int ret;

	/*
	 * Fix CVE-2016-3134
	 *  -ret = kgr_check_entry(e, name);
	 *  -if (ret)
	 *  -	return ret;
	 *  -
	 */
	t = arpt_get_target(e);
	target = kgr_xt_request_find_target(NFPROTO_ARP, t->u.user.name,
					t->u.user.revision);
	if (IS_ERR(target)) {
		duprintf("find_check_entry: `%s' not found\n", t->u.user.name);
		ret = PTR_ERR(target);
		goto out;
	}
	t->u.kernel.target = target;

	ret = kgr_check_target(e, name);
	if (ret)
		goto err;
	return 0;
err:
	module_put(t->u.kernel.target->me);
out:
	return ret;
}

/* inlined */
static inline void kgr_cleanup_entry(struct arpt_entry *e)
{
	struct xt_tgdtor_param par;
	struct xt_entry_target *t;

	t = arpt_get_target(e);
	par.target   = t->u.kernel.target;
	par.targinfo = t->data;
	par.family   = NFPROTO_ARP;
	if (par.target->destroy != NULL)
		par.target->destroy(&par);
	module_put(par.target->me);
}

/* to-be-patched function, but ambiguous */
int kgr_translate_table(struct xt_table_info *newinfo, void *entry0,
	const struct arpt_replace *repl)
{
	struct arpt_entry *iter;
	/*
	 * Fix CVE-2016-4997
	 *  +next line
	 */
	unsigned int *offsets;
	unsigned int i;
	int ret = 0;

	newinfo->size = repl->size;
	newinfo->number = repl->num_entries;

	/* Init all hooks to impossible value. */
	for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
		newinfo->hook_entry[i] = 0xFFFFFFFF;
		newinfo->underflow[i] = 0xFFFFFFFF;
	}

	duprintf("translate_table: size %u\n", newinfo->size);
	/*
	 * Fix CVE-2016-4997
	 *  +next 3 lines
	 */
	offsets = kgr_xt_alloc_entry_offsets(newinfo->number);
	if (!offsets)
		return -ENOMEM;
	i = 0;

	/* Walk through entries, checking offsets. */
	xt_entry_foreach(iter, entry0, newinfo->size) {
		ret = kgr_check_entry_size_and_hooks(iter, newinfo, entry0,
						 entry0 + repl->size,
						 repl->hook_entry,
						 repl->underflow,
						 repl->valid_hooks);
		if (ret != 0)
		/*
		 * Fix CVE-2016-4997
		 *  -	break;
		 *  +next 3 lines
		 */
			goto out_free;
		if (i < repl->num_entries)
			offsets[i] = (void *)iter - entry0;
		++i;
		if (strcmp(arpt_get_target(iter)->u.user.name,
		    XT_ERROR_TARGET) == 0)
			++newinfo->stacksize;
	}
	duprintf("translate_table: ARPT_ENTRY_ITERATE gives %d\n", ret);
	if (ret != 0)
	/*
	 * Fix CVE-2016-4997
	 *  -	return ret;
	 *  +next 3 lines
	 */
		goto out_free;

	ret = -EINVAL;
	if (i != repl->num_entries) {
		duprintf("translate_table: %u not %u entries\n",
			 i, repl->num_entries);
		/*
		 * Fix CVE-2016-4997
		 *  -return -EINVAL;
		 *  +next line
		 */
		goto out_free;
	}

	/* Check hooks all assigned */
	for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
		/* Only hooks which are valid */
		if (!(repl->valid_hooks & (1 << i)))
			continue;
		if (newinfo->hook_entry[i] == 0xFFFFFFFF) {
			duprintf("Invalid hook entry %u %u\n",
				 i, repl->hook_entry[i]);
			/*
			 * Fix CVE-2016-4997
			 *  -return -EINVAL;
			 *  +next line
			 */
			goto out_free;
		}
		if (newinfo->underflow[i] == 0xFFFFFFFF) {
			duprintf("Invalid underflow %u %u\n",
				 i, repl->underflow[i]);
			/*
			 * Fix CVE-2016-4997
			 *  -return -EINVAL;
			 *  +next line
			 */
			goto out_free;
		}
	}

	/*
	 * Fix CVE-2016-4997
	 *  -if (!kgr_mark_source_chains(newinfo, repl->valid_hooks, entry0)) {
	 *  -	duprintf("Looping hook\n");
	 *  -	return -ELOOP;
	 *  -}
	 *  +next 5 lines
	 */
	if (!kgr_mark_source_chains(newinfo, repl->valid_hooks, entry0, offsets)) {
		ret = -ELOOP;
		goto out_free;
	}
	kvfree(offsets);

	/* Finally, each sanity check must pass */
	i = 0;
	xt_entry_foreach(iter, entry0, newinfo->size) {
		ret = kgr_find_check_entry(iter, repl->name, repl->size);
		if (ret != 0)
			break;
		++i;
	}

	if (ret != 0) {
		xt_entry_foreach(iter, entry0, newinfo->size) {
			if (i-- == 0)
				break;
			kgr_cleanup_entry(iter);
		}
		return ret;
	}

	/* And one copy for every other CPU */
	for_each_possible_cpu(i) {
		if (newinfo->entries[i] && newinfo->entries[i] != entry0)
			memcpy(newinfo->entries[i], entry0, newinfo->size);
	}

	return ret;
	/*
	 * Fix CVE-2016-4997
	 *  +3 lines
	 */
out_free:
	kvfree(offsets);
	return ret;
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 *  -static inline int
 */
static int
kgr_check_compat_entry_size_and_hooks(struct compat_arpt_entry *e,
				  struct xt_table_info *newinfo,
				  unsigned int *size,
				  const unsigned char *base,
/*
 *				  const unsigned char *limit,
 *				  const unsigned int *hook_entries,
 *				  const unsigned int *underflows,
 *				  const char *name)
 */
				  const unsigned char *limit)
{
	struct xt_entry_target *t;
	struct xt_target *target;
	unsigned int entry_offset;
	/*
	 * Fix CVE-2016-4997
	 *  -int ret, off, h;
	 *  +1 line
	 */
	int ret, off;

	duprintf("check_compat_entry_size_and_hooks %p\n", e);
	if ((unsigned long)e % __alignof__(struct compat_arpt_entry) != 0 ||
	/*
	 * Fix CVE-2016-3134
	 *  -    (unsigned char *)e + sizeof(struct compat_arpt_entry) >= limit) {
	 *  +next 2 rows
	 */
	    (unsigned char *)e + sizeof(struct compat_arpt_entry) >= limit ||
	    (unsigned char *)e + e->next_offset > limit) {
		duprintf("Bad offset %p, limit = %p\n", e, limit);
		return -EINVAL;
	}

	if (e->next_offset < sizeof(struct compat_arpt_entry) +
			     sizeof(struct compat_xt_entry_target)) {
		duprintf("checking: element %p size %u\n",
			 e, e->next_offset);
		return -EINVAL;
	}

	/*
	 * Fix CVE-2016-3134
	 *  -ret = kgr_check_entry((struct arpt_entry *)e, name);
	 *  +next row
	 */
	/*
	 * Fix CVE-2016-4997
	 *  -* For purposes of check_entry casting the compat entry is fine
	 *  -ret = kgr_check_entry((struct arpt_entry *)e);
	 *  +5 lines
	 */
	if (!kgr_arp_checkentry(&e->arp))
		return -EINVAL;

	ret = kgr_xt_compat_check_entry_offsets(e, e->elems, e->target_offset,
					    e->next_offset);
	if (ret)
		return ret;

	off = sizeof(struct arpt_entry) - sizeof(struct compat_arpt_entry);
	entry_offset = (void *)e - (void *)base;

	t = compat_arpt_get_target(e);
	target = kgr_xt_request_find_target(NFPROTO_ARP, t->u.user.name,
					t->u.user.revision);
	if (IS_ERR(target)) {
		duprintf("check_compat_entry_size_and_hooks: `%s' not found\n",
			 t->u.user.name);
		ret = PTR_ERR(target);
		goto out;
	}
	t->u.kernel.target = target;

	off += kgr_xt_compat_target_offset(target);
	*size += off;
	ret = kgr_xt_compat_add_offset(NFPROTO_ARP, entry_offset, off);
	if (ret)
		goto release_target;

	/*
	 * Fix CVE-2016-4997
	 *  -* Check hooks & underflows
	 *  -for (h = 0; h < NF_ARP_NUMHOOKS; h++) {
	 *  -	if ((unsigned char *)e - base == hook_entries[h])
	 *  -		newinfo->hook_entry[h] = hook_entries[h];
	 *  -	if ((unsigned char *)e - base == underflows[h])
	 *  -		newinfo->underflow[h] = underflows[h];
	 *  -}
	 *  -
	 *  -* Clear counters and comefrom
	 *  -memset(&e->counters, 0, sizeof(e->counters));
	 *  -e->comefrom = 0;
	 */
	return 0;

release_target:
	module_put(t->u.kernel.target->me);
out:
	return ret;
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 *  -static int
 */
static void
kgr_compat_copy_entry_from_user(struct compat_arpt_entry *e, void **dstptr,
/*
 *			    unsigned int *size, const char *name,
 */
			    unsigned int *size,
			    struct xt_table_info *newinfo, unsigned char *base)
{
	struct xt_entry_target *t;
	struct xt_target *target;
	struct arpt_entry *de;
	unsigned int origsize;
	/*
	 * Fix CVE-2016-4997
	 *  -int ret, h;
	 */
	int h;

	/*
	 * Fix CVE-2016-4997
	 *  -ret = 0;
	 */
	origsize = *size;
	de = (struct arpt_entry *)*dstptr;
	memcpy(de, e, sizeof(struct arpt_entry));
	memcpy(&de->counters, &e->counters, sizeof(e->counters));

	*dstptr += sizeof(struct arpt_entry);
	*size += sizeof(struct arpt_entry) - sizeof(struct compat_arpt_entry);

	de->target_offset = e->target_offset - (origsize - *size);
	t = compat_arpt_get_target(e);
	target = t->u.kernel.target;
	kgr_xt_compat_target_from_user(t, dstptr, size);

	de->next_offset = e->next_offset - (origsize - *size);
	for (h = 0; h < NF_ARP_NUMHOOKS; h++) {
		if ((unsigned char *)de - base < newinfo->hook_entry[h])
			newinfo->hook_entry[h] -= origsize - *size;
		if ((unsigned char *)de - base < newinfo->underflow[h])
			newinfo->underflow[h] -= origsize - *size;
	}
	/*
	 * Fix CVE-2016-4997
	 *  -return ret;
	 */
}

/* inlined */
static inline void kgr_compat_release_entry(struct compat_arpt_entry *e)
{
	struct xt_entry_target *t;

	t = compat_arpt_get_target(e);
	module_put(t->u.kernel.target->me);
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 *  -static int kgr_translate_compat_table(const char *name,
 *  -				  unsigned int valid_hooks,
 *  -				  struct xt_table_info **pinfo,
 */
static int kgr_translate_compat_table(struct xt_table_info **pinfo,
				  void **pentry0,
/*
 *				  unsigned int total_size,
 *				  unsigned int number,
 *				  unsigned int *hook_entries,
 *				  unsigned int *underflows)
 */
				  const struct compat_arpt_replace *compatr)
{
	unsigned int i, j;
	struct xt_table_info *newinfo, *info;
	void *pos, *entry0, *entry1;
	struct compat_arpt_entry *iter0;
	/*
	 * Fix CVE-2016-4997
	 *  -struct arpt_entry *iter1;
	 *  +1 line
	 */
	struct arpt_replace repl;
	unsigned int size;
	int ret = 0;

	info = *pinfo;
	entry0 = *pentry0;
	/*
	 * Fix CVE-2016-4997
	 *  -size = total_size;
	 *  -info->number = number;
	 *  -
	 *  -* Init all hooks to impossible value.
	 *  -for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
	 *  -	info->hook_entry[i] = 0xFFFFFFFF;
	 *  -	info->underflow[i] = 0xFFFFFFFF;
	 *  -}
	 *  +2 lines
	 */
	size = compatr->size;
	info->number = compatr->num_entries;

	duprintf("translate_compat_table: size %u\n", info->size);
	j = 0;
	kgr_xt_compat_lock(NFPROTO_ARP);
	/*
	 * Fix CVE-2016-4997
	 *  -kgr_xt_compat_init_offsets(NFPROTO_ARP, number);
	 *  +1 line
	 */
	kgr_xt_compat_init_offsets(NFPROTO_ARP, compatr->num_entries);
	/* Walk through entries, checking offsets. */
	/*
	 * Fix CVE-2016-4997
	 *  -xt_entry_foreach(iter0, entry0, total_size) {
	 *  +1 line
	 */
	xt_entry_foreach(iter0, entry0, compatr->size) {
		ret = kgr_check_compat_entry_size_and_hooks(iter0, info, &size,
							entry0,
							/*
							 * Fix CVE-2016-4997
							 *  -entry0 + total_size,
							 *  -hook_entries,
							 *  -underflows,
							 *  -name);
							 *  +1 line
							 */
							entry0 + compatr->size);
		if (ret != 0)
			goto out_unlock;
		++j;
	}

	ret = -EINVAL;
	/*
	 * Fix CVE-2016-4997
	 *  -if (j != number) {
	 *  +1 line
	 */
	if (j != compatr->num_entries) {
		duprintf("translate_compat_table: %u not %u entries\n",
			/*
			 * Fix CVE-2016-4997
			 *  -j, number);
			 *  +1 line
			 */
			 j, compatr->num_entries);
		goto out_unlock;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -* Check hooks all assigned
	 *  -for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
	 *  -	* Only hooks which are valid
	 *  -	if (!(valid_hooks & (1 << i)))
	 *  -		continue;
	 *  -	if (info->hook_entry[i] == 0xFFFFFFFF) {
	 *  -		duprintf("Invalid hook entry %u %u\n",
	 *  -			 i, hook_entries[i]);
	 *  -		goto out_unlock;
	 *  -	}
	 *  -	if (info->underflow[i] == 0xFFFFFFFF) {
	 *  -		duprintf("Invalid underflow %u %u\n",
	 *  -			 i, underflows[i]);
	 *  -		goto out_unlock;
	 *  -	}
	 *  -}
	 *  -
	 */
	ret = -ENOMEM;
	newinfo = kgr_xt_alloc_table_info(size);
	if (!newinfo)
		goto out_unlock;

	/*
	 * Fix CVE-2016-4997
	 *  -newinfo->number = number;
	 *  +1 line
	 */
	newinfo->number = compatr->num_entries;
	for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
		newinfo->hook_entry[i] = info->hook_entry[i];
		newinfo->underflow[i] = info->underflow[i];
	}
	entry1 = newinfo->entries[raw_smp_processor_id()];
	pos = entry1;
	/*
	 * Fix CVE-2016-4997
	 *  -size = total_size;
	 *  -xt_entry_foreach(iter0, entry0, total_size) {
	 *  -	ret = kgr_compat_copy_entry_from_user(iter0, &pos, &size,
	 *  -					  name, newinfo, entry1);
	 *  -	if (ret != 0)
	 *  -		break;
	 *  -}
	 *  +7 lines
	 */
	size = compatr->size;
	xt_entry_foreach(iter0, entry0, compatr->size)
		kgr_compat_copy_entry_from_user(iter0, &pos, &size,
					    newinfo, entry1);

	/* all module references in entry0 are now gone */

	kgr_xt_compat_flush_offsets(NFPROTO_ARP);
	kgr_xt_compat_unlock(NFPROTO_ARP);
	/*
	 * Fix CVE-2016-4997
	 *  -if (ret)
	 *  -	goto free_newinfo;
	 *  -
	 *  -ret = -ELOOP;
	 *  -if (!kgr_mark_source_chains(newinfo, valid_hooks, entry1))
	 *  -	goto free_newinfo;
	 *  +1 line
	 */
	memcpy(&repl, compatr, sizeof(*compatr));

	/*
	 * Fix CVE-2016-4997
	 *  -i = 0;
	 *  -xt_entry_foreach(iter1, entry1, newinfo->size) {
	 *  -	ret = kgr_check_target(iter1, name);
	 *  -	if (ret != 0)
	 *  -		break;
	 *  -	++i;
	 *  -	if (strcmp(arpt_get_target(iter1)->u.user.name,
	 *  -	    XT_ERROR_TARGET) == 0)
	 *  -		++newinfo->stacksize;
	 *  -}
	 *  -if (ret) {
	 *  -	 *
	 *  -	 * The first i matches need cleanup_entry (calls ->destroy)
	 *  -	 * because they had called ->check already. The other j-i
	 *  -	 * entries need only release.
	 *  -	 *
	 *  -	int skip = i;
	 *  -	j -= i;
	 *  -	xt_entry_foreach(iter0, entry0, newinfo->size) {
	 *  -		if (skip-- > 0)
	 *  -			continue;
	 *  -		if (j-- == 0)
	 *  -			break;
	 *  -		kgr_compat_release_entry(iter0);
	 *  -	}
	 *  -	xt_entry_foreach(iter1, entry1, newinfo->size) {
	 *  -		if (i-- == 0)
	 *  -			break;
	 *  -		kgr_cleanup_entry(iter1);
	 *  -	}
	 *  -	kgr_xt_free_table_info(newinfo);
	 *  -	return ret;
	 *  +3 lines
	 */
	for (i = 0; i < NF_ARP_NUMHOOKS; i++) {
		repl.hook_entry[i] = newinfo->hook_entry[i];
		repl.underflow[i] = newinfo->underflow[i];
	}

	/*
	 * Fix CVE-2016-4997
	 *  +7 lines
	 */
	repl.num_counters = 0;
	repl.counters = NULL;
	repl.size = newinfo->size;
	ret = kgr_translate_table(newinfo, entry1, &repl);
	if (ret)
		goto free_newinfo;

	/* And one copy for every other CPU */
	for_each_possible_cpu(i)
		if (newinfo->entries[i] && newinfo->entries[i] != entry1)
			memcpy(newinfo->entries[i], entry1, newinfo->size);

	*pinfo = newinfo;
	*pentry0 = entry1;
	kgr_xt_free_table_info(info);
	return 0;

free_newinfo:
	kgr_xt_free_table_info(newinfo);
/*
 * Fix CVE-2016-4997
 *  -out:
 *  -	xt_entry_foreach(iter0, entry0, total_size) {
 *  +5 lines
 */
	return ret;
out_unlock:
	kgr_xt_compat_flush_offsets(NFPROTO_ARP);
	kgr_xt_compat_unlock(NFPROTO_ARP);
	xt_entry_foreach(iter0, entry0, compatr->size) {
		if (j-- == 0)
			break;
		kgr_compat_release_entry(iter0);
	}
	return ret;
/*
 * Fix CVE-2016-4997
 *  -out_unlock:
 *  -	kgr_xt_compat_flush_offsets(NFPROTO_ARP);
 *  -	kgr_xt_compat_unlock(NFPROTO_ARP);
 *  -	goto out;
 */
}

/* inlined */
static int kgr_compat_do_replace(struct net *net, void __user *user,
			     unsigned int len)
{
	int ret;
	struct compat_arpt_replace tmp;
	struct xt_table_info *newinfo;
	void *loc_cpu_entry;
	struct arpt_entry *iter;

	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return -EFAULT;

	/* overflow check */
	if (tmp.size >= INT_MAX / num_possible_cpus())
		return -ENOMEM;
	if (tmp.num_counters >= INT_MAX / sizeof(struct xt_counters))
		return -ENOMEM;
	tmp.name[sizeof(tmp.name)-1] = 0;

	newinfo = kgr_xt_alloc_table_info(tmp.size);
	if (!newinfo)
		return -ENOMEM;

	/* choose the copy that is on our node/cpu */
	loc_cpu_entry = newinfo->entries[raw_smp_processor_id()];
	if (copy_from_user(loc_cpu_entry, user + sizeof(tmp), tmp.size) != 0) {
		ret = -EFAULT;
		goto free_newinfo;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -ret = kgr_translate_compat_table(tmp.name, tmp.valid_hooks,
	 *  -			     &newinfo, &loc_cpu_entry, tmp.size,
	 *  -			     tmp.num_entries, tmp.hook_entry,
	 *  -			     tmp.underflow);
	 *  +1 line
	 */
	ret = kgr_translate_compat_table(&newinfo, &loc_cpu_entry, &tmp);
	if (ret != 0)
		goto free_newinfo;

	duprintf("compat_do_replace: Translated table\n");

	ret = kgr__do_replace(net, tmp.name, tmp.valid_hooks, newinfo,
			   tmp.num_counters, compat_ptr(tmp.counters));
	if (ret)
		goto free_newinfo_untrans;
	return 0;

 free_newinfo_untrans:
	xt_entry_foreach(iter, loc_cpu_entry, newinfo->size)
		kgr_cleanup_entry(iter);
 free_newinfo:
	kgr_xt_free_table_info(newinfo);
	return ret;
}

/* patched function */
int kgr_compat_do_arpt_set_ctl(struct sock *sk, int cmd, void __user *user,
	unsigned int len)
{
	int ret;

	if (!ns_capable(sock_net(sk)->user_ns, CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case ARPT_SO_SET_REPLACE:
		ret = kgr_compat_do_replace(sock_net(sk), user, len);
		break;

	case ARPT_SO_SET_ADD_COUNTERS:
		ret = kgr_do_add_counters(sock_net(sk), user, len, 1);
		break;

	default:
		duprintf("do_arpt_set_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

/* inlined */
static inline struct arpt_entry *
kgr_get_entry(const void *base, unsigned int offset)
{
	return (struct arpt_entry *)(base + offset);
}

/* inlined */
static inline int kgr_arp_devaddr_compare(const struct arpt_devaddr_info *ap,
				      const char *hdr_addr, int len)
{
	int i, ret;

	if (len > ARPT_DEV_ADDR_LEN_MAX)
		len = ARPT_DEV_ADDR_LEN_MAX;

	ret = 0;
	for (i = 0; i < len; i++)
		ret |= (hdr_addr[i] ^ ap->addr[i]) & ap->mask[i];

	return ret != 0;
}

/* inlined */
static unsigned long kgr_ifname_compare(const char *_a, const char *_b, const char *_mask)
{
#ifdef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	unsigned long ret = ifname_compare_aligned(_a, _b, _mask);
#else
	unsigned long ret = 0;
	const u16 *a = (const u16 *)_a;
	const u16 *b = (const u16 *)_b;
	const u16 *mask = (const u16 *)_mask;
	int i;

	for (i = 0; i < IFNAMSIZ/sizeof(u16); i++)
		ret |= (a[i] ^ b[i]) & mask[i];
#endif
	return ret;
}

/* inlined */
static inline int kgr_arp_packet_match(const struct arphdr *arphdr,
				   struct net_device *dev,
				   const char *indev,
				   const char *outdev,
				   const struct arpt_arp *arpinfo)
{
	const char *arpptr = (char *)(arphdr + 1);
	const char *src_devaddr, *tgt_devaddr;
	__be32 src_ipaddr, tgt_ipaddr;
	long ret;

#define FWINV(bool, invflg) ((bool) ^ !!(arpinfo->invflags & (invflg)))

	if (FWINV((arphdr->ar_op & arpinfo->arpop_mask) != arpinfo->arpop,
		  ARPT_INV_ARPOP)) {
		dprintf("ARP operation field mismatch.\n");
		dprintf("ar_op: %04x info->arpop: %04x info->arpop_mask: %04x\n",
			arphdr->ar_op, arpinfo->arpop, arpinfo->arpop_mask);
		return 0;
	}

	if (FWINV((arphdr->ar_hrd & arpinfo->arhrd_mask) != arpinfo->arhrd,
		  ARPT_INV_ARPHRD)) {
		dprintf("ARP hardware address format mismatch.\n");
		dprintf("ar_hrd: %04x info->arhrd: %04x info->arhrd_mask: %04x\n",
			arphdr->ar_hrd, arpinfo->arhrd, arpinfo->arhrd_mask);
		return 0;
	}

	if (FWINV((arphdr->ar_pro & arpinfo->arpro_mask) != arpinfo->arpro,
		  ARPT_INV_ARPPRO)) {
		dprintf("ARP protocol address format mismatch.\n");
		dprintf("ar_pro: %04x info->arpro: %04x info->arpro_mask: %04x\n",
			arphdr->ar_pro, arpinfo->arpro, arpinfo->arpro_mask);
		return 0;
	}

	if (FWINV((arphdr->ar_hln & arpinfo->arhln_mask) != arpinfo->arhln,
		  ARPT_INV_ARPHLN)) {
		dprintf("ARP hardware address length mismatch.\n");
		dprintf("ar_hln: %02x info->arhln: %02x info->arhln_mask: %02x\n",
			arphdr->ar_hln, arpinfo->arhln, arpinfo->arhln_mask);
		return 0;
	}

	src_devaddr = arpptr;
	arpptr += dev->addr_len;
	memcpy(&src_ipaddr, arpptr, sizeof(u32));
	arpptr += sizeof(u32);
	tgt_devaddr = arpptr;
	arpptr += dev->addr_len;
	memcpy(&tgt_ipaddr, arpptr, sizeof(u32));

	if (FWINV(kgr_arp_devaddr_compare(&arpinfo->src_devaddr, src_devaddr, dev->addr_len),
		  ARPT_INV_SRCDEVADDR) ||
	    FWINV(kgr_arp_devaddr_compare(&arpinfo->tgt_devaddr, tgt_devaddr, dev->addr_len),
		  ARPT_INV_TGTDEVADDR)) {
		dprintf("Source or target device address mismatch.\n");

		return 0;
	}

	if (FWINV((src_ipaddr & arpinfo->smsk.s_addr) != arpinfo->src.s_addr,
		  ARPT_INV_SRCIP) ||
	    FWINV(((tgt_ipaddr & arpinfo->tmsk.s_addr) != arpinfo->tgt.s_addr),
		  ARPT_INV_TGTIP)) {
		dprintf("Source or target IP address mismatch.\n");

		dprintf("SRC: %pI4. Mask: %pI4. Target: %pI4.%s\n",
			&src_ipaddr,
			&arpinfo->smsk.s_addr,
			&arpinfo->src.s_addr,
			arpinfo->invflags & ARPT_INV_SRCIP ? " (INV)" : "");
		dprintf("TGT: %pI4 Mask: %pI4 Target: %pI4.%s\n",
			&tgt_ipaddr,
			&arpinfo->tmsk.s_addr,
			&arpinfo->tgt.s_addr,
			arpinfo->invflags & ARPT_INV_TGTIP ? " (INV)" : "");
		return 0;
	}

	/* Look for ifname matches.  */
	ret = kgr_ifname_compare(indev, arpinfo->iniface, arpinfo->iniface_mask);

	if (FWINV(ret != 0, ARPT_INV_VIA_IN)) {
		dprintf("VIA in mismatch (%s vs %s).%s\n",
			indev, arpinfo->iniface,
			arpinfo->invflags&ARPT_INV_VIA_IN ?" (INV)":"");
		return 0;
	}

	ret = kgr_ifname_compare(outdev, arpinfo->outiface, arpinfo->outiface_mask);

	if (FWINV(ret != 0, ARPT_INV_VIA_OUT)) {
		dprintf("VIA out mismatch (%s vs %s).%s\n",
			outdev, arpinfo->outiface,
			arpinfo->invflags&ARPT_INV_VIA_OUT ?" (INV)":"");
		return 0;
	}

	return 1;
#undef FWINV
}

/* inlined */
static inline __pure
struct arpt_entry *kgr_arpt_next_entry(const struct arpt_entry *entry)
{
	return (void *)entry + entry->next_offset;
}

/* inlined, percpu variable */
static inline unsigned int kgr_xt_write_recseq_begin(void)
{
	unsigned int addend;

	/*
	 * Low order bit of sequence is set if we already
	 * called xt_write_recseq_begin().
	 */
	addend = (__this_cpu_read(kgr_xt_recseq->sequence) + 1) & 1;

	/*
	 * This is kind of a write_seqcount_begin(), but addend is 0 or 1
	 * We dont check addend value to avoid a test and conditional jump,
	 * since addend is most likely 1
	 */
	__this_cpu_add(kgr_xt_recseq->sequence, addend);
	smp_wmb();

	return addend;
}

/* inlined, percpu variable */
static inline void kgr_xt_write_recseq_end(unsigned int addend)
{
	/* this is kind of a write_seqcount_end(), but addend is 0 or 1 */
	smp_wmb();
	__this_cpu_add(kgr_xt_recseq->sequence, addend);
}

/* patched function */
unsigned int kgr_arpt_do_table(struct sk_buff *skb, unsigned int hook,
	const struct net_device *in, const struct net_device *out,
	struct xt_table *table)
{
	static const char nulldevname[IFNAMSIZ] __attribute__((aligned(sizeof(long))));
	unsigned int verdict = NF_DROP;
	const struct arphdr *arp;
	struct arpt_entry *e, *back;
	const char *indev, *outdev;
	void *table_base;
	const struct xt_table_info *private;
	struct xt_action_param acpar;
	unsigned int addend;

	if (!pskb_may_pull(skb, arp_hdr_len(skb->dev)))
		return NF_DROP;

	indev = in ? in->name : nulldevname;
	outdev = out ? out->name : nulldevname;

	local_bh_disable();
	addend = kgr_xt_write_recseq_begin();
	private = table->private;
	table_base = private->entries[smp_processor_id()];

	e = kgr_get_entry(table_base, private->hook_entry[hook]);
	back = kgr_get_entry(table_base, private->underflow[hook]);

	acpar.in      = in;
	acpar.out     = out;
	acpar.hooknum = hook;
	acpar.family  = NFPROTO_ARP;
	acpar.hotdrop = false;

	arp = arp_hdr(skb);
	do {
		const struct xt_entry_target *t;

		if (!kgr_arp_packet_match(arp, skb->dev, indev, outdev, &e->arp)) {
			e = kgr_arpt_next_entry(e);
			continue;
		}

		ADD_COUNTER(e->counters, arp_hdr_len(skb->dev), 1);

		t = kgr_arpt_get_target_c(e);

		/* Standard target? */
		if (!t->u.kernel.target->target) {
			int v;

			v = ((struct xt_standard_target *)t)->verdict;
			if (v < 0) {
				/* Pop from stack? */
				if (v != XT_RETURN) {
					verdict = (unsigned int)(-v) - 1;
					break;
				}
				e = back;
				back = kgr_get_entry(table_base, back->comefrom);
				continue;
			}
			if (table_base + v
			    != kgr_arpt_next_entry(e)) {
				/* Save old back ptr in next entry */
				struct arpt_entry *next = kgr_arpt_next_entry(e);
				next->comefrom = (void *)back - table_base;

				/* set back pointer to next entry */
				back = next;
			}

			e = kgr_get_entry(table_base, v);
			continue;
		}

		/* Targets which reenter must return
		 * abs. verdicts
		 */
		acpar.target   = t->u.kernel.target;
		acpar.targinfo = t->data;
		verdict = t->u.kernel.target->target(skb, &acpar);

		/* Target might have changed stuff. */
		arp = arp_hdr(skb);

		if (verdict == XT_CONTINUE)
			e = kgr_arpt_next_entry(e);
		else
			/* Verdict */
			break;
	} while (!acpar.hotdrop);
	kgr_xt_write_recseq_end(addend);
	local_bh_enable();

	if (acpar.hotdrop)
		return NF_DROP;
	else
		return verdict;
}

/* inlined */
static int kgr_do_replace(struct net *net, const void __user *user,
	unsigned int len)
{
	int ret;
	struct arpt_replace tmp;
	struct xt_table_info *newinfo;
	void *loc_cpu_entry;
	struct arpt_entry *iter;

	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return -EFAULT;

	/* overflow check */
	if (tmp.num_counters >= INT_MAX / sizeof(struct xt_counters))
		return -ENOMEM;
	tmp.name[sizeof(tmp.name)-1] = 0;

	newinfo = kgr_xt_alloc_table_info(tmp.size);
	if (!newinfo)
		return -ENOMEM;

	/* choose the copy that is on our node/cpu */
	loc_cpu_entry = newinfo->entries[raw_smp_processor_id()];
	if (copy_from_user(loc_cpu_entry, user + sizeof(tmp),
			   tmp.size) != 0) {
		ret = -EFAULT;
		goto free_newinfo;
	}

	ret = kgr_translate_table(newinfo, loc_cpu_entry, &tmp);
	if (ret != 0)
		goto free_newinfo;

	duprintf("arp_tables: Translated table\n");

	ret = kgr__do_replace(net, tmp.name, tmp.valid_hooks, newinfo,
			   tmp.num_counters, tmp.counters);
	if (ret)
		goto free_newinfo_untrans;
	return 0;

 free_newinfo_untrans:
	xt_entry_foreach(iter, loc_cpu_entry, newinfo->size)
		kgr_cleanup_entry(iter);
 free_newinfo:
	kgr_xt_free_table_info(newinfo);
	return ret;
}

/* patched function */
int kgr_do_arpt_set_ctl(struct sock *sk, int cmd, void __user *user,
	unsigned int len)
{
	int ret;

	if (!ns_capable(sock_net(sk)->user_ns, CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case ARPT_SO_SET_REPLACE:
		ret = kgr_do_replace(sock_net(sk), user, len);
		break;

	case ARPT_SO_SET_ADD_COUNTERS:
		ret = kgr_do_add_counters(sock_net(sk), user, len, 0);
		break;

	default:
		duprintf("do_arpt_set_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

/* patched function */
struct xt_table *kgr_arpt_register_table(struct net *net,
	const struct xt_table *table, const struct arpt_replace *repl)
{
	int ret;
	struct xt_table_info *newinfo;
	struct xt_table_info bootstrap = {0};
	void *loc_cpu_entry;
	struct xt_table *new_table;

	newinfo = kgr_xt_alloc_table_info(repl->size);
	if (!newinfo) {
		ret = -ENOMEM;
		goto out;
	}

	/* choose the copy on our node/cpu */
	loc_cpu_entry = newinfo->entries[raw_smp_processor_id()];
	memcpy(loc_cpu_entry, repl->entries, repl->size);

	ret = kgr_translate_table(newinfo, loc_cpu_entry, repl);
	duprintf("arpt_register_table: translate table gives %d\n", ret);
	if (ret != 0)
		goto out_free;

	new_table = kgr_xt_register_table(net, table, &bootstrap, newinfo);
	if (IS_ERR(new_table)) {
		ret = PTR_ERR(new_table);
		goto out_free;
	}
	return new_table;

out_free:
	kgr_xt_free_table_info(newinfo);
out:
	return ERR_PTR(ret);
}

static int kgr_patch_bsc971793_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc971793_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc971793_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc971793_module_nb = {
	.notifier_call = kgr_patch_bsc971793_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc971793_arp_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc971793_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc971793_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc971793_arp_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc971793_module_nb);
}
